package HTret;
import java.net.*;
import java.io.*;
import java.util.*;

class Engine implements Runnable {
  int threadNo;
  String objective;

  static String fileSeparator = System.getProperty("file.separator");

  static String constHREF = "href=";
  static String constSRC  = "src=";
  static String constBACKGROUND = "background=";
  static String DEFAULTFILENAME = "index.html";

  static int BUFMAX = 16;

  public Engine(int threadNo, String url)  {
    this.threadNo = threadNo;
    objective = url;
  }

  public void run() {
    URL  url = null;
    InputStream is;
    FileOutputStream fos;
    String defaultSupplementedFileName;

    try {
      url = new URL(objective);
    } catch(MalformedURLException e) {
      ControlThread.g.log("# Exception!(MalformedURL) " + url.toString());
      return;
    }
    ControlThread.g.monitor(threadNo, "[-]:" + objective);

    // Get filename
    String fileName = url.getFile();
    fileName = fileName.substring(1);

    // Get directoryname
    String dirString;
    int w = fileName.lastIndexOf('/');
    if (w != -1)  {
      dirString = fileName.substring(0, w);
    } else {
      dirString = "";
    }

    // Make directory name appropriate by changing delimiters
    dirString = ControlThread.g.getLocalPath() + dirString;
    StringBuffer dirName = new StringBuffer(dirString.length());
    for(int i = 0; i < dirString.length(); ++i)  {
      char ch = dirString.charAt(i);
      if (ch == '/')  {
        dirName.append(fileSeparator);
      } else {
        dirName.append(ch);
      }
    }
    new File(dirName.toString()).mkdirs();

    defaultSupplementedFileName = defaultSupplement(fileName);
    File check = new File(defaultSupplementedFileName);

    if (new File(defaultSupplementedFileName).exists()) {
      if (isHTML(defaultSupplementedFileName))  {
        ControlThread.g.monitor(threadNo, "[P]:" + defaultSupplementedFileName);
        try {
          parseDocument(new FileInputStream(defaultSupplementedFileName),
                        null, url, defaultSupplementedFileName);
        } catch(FileNotFoundException e) {}    // Never occurs.
      }
    } else {                  // download document.
      boolean success = false;
      for (int i = 0; (i < 3) && (!success); i++)  {
        try {
          switch (i) {
            case 0 :
              ControlThread.g.monitor(threadNo, "[D]:" + objective);
              break;
            case 1 :
              ControlThread.g.monitor(threadNo, "[?]:" + objective);
              break;
            default:
              ControlThread.g.monitor(threadNo, "[!]:" + objective);
          }
          URLConnection uc = url.openConnection();
          is = uc.getInputStream();
          fos = new FileOutputStream(defaultSupplementedFileName);
          if  (isHTML(defaultSupplementedFileName))  {
            success = parseDocument(is, fos, url, defaultSupplementedFileName);
          } else {
            success = downloadDocument(is, fos, url, defaultSupplementedFileName);
          }
        } catch(IOException e) {
          success = false;
        }
      }
      if  (!success)  ControlThread.g.log("# Exception! " + url.toString());
    }


    ControlThread.g.monitor(threadNo, "    -----");

  }

  public boolean downloadDocument(InputStream is, FileOutputStream fos, URL url, String defaultSupplementedFileName)  {
    boolean success = true;
    int ch;

    try  {
      while ((ch = is.read()) != -1) {
        fos.write(ch);
      }
      is.close();
      fos.close();
    } catch(IOException e) {
      try {
        is.close();
        if (fos != null)  fos.close();
      } catch(Exception ee) {}
      try {
        new File(defaultSupplementedFileName).delete();
        success = false;
      } catch(Exception ee) {}
      ControlThread.g.log("# Exception! " + url.toString());
    }
    return success;
  }

  public boolean parseDocument(InputStream is, FileOutputStream fos, URL url, String defaultSupplementedFileName) {
    boolean success = true;
    int ch;

    try {
      char[] workArea = new char[BUFMAX];
      int currentPosition = BUFMAX;

      boolean inProgress = false;
      int quoteCount = 0;
      StringBuffer urlWork = new StringBuffer();

      boolean inTag = false;
      while ((ch = is.read()) != -1) {
        if  (++currentPosition >= BUFMAX) currentPosition = 0;
        workArea[currentPosition] = (char)ch;
        switch (ch) {
          case '<' :
            inTag = true;
            quoteCount = 0;
            break;
          case '>' :
          case '#' :
            inTag = false;
            inProgress = false;
            registerURL(url, urlWork.toString(), fos);
            urlWork = new StringBuffer();
            break;
          case ' ' :
            if (urlWork.length() > 0)  {
              inProgress = false;
              registerURL(url, urlWork.toString(), fos);
              urlWork = new StringBuffer();
            }
            break;
          case '"' :
            if  ((++quoteCount % 2) == 0)  {
              inProgress = false;
              registerURL(url, urlWork.toString(), fos);
              urlWork = new StringBuffer();
            }
            break;
          case '=' :
            if  (inTag)  {
              if  (isAnchor(workArea, currentPosition, constHREF) ||
                   isAnchor(workArea, currentPosition, constSRC)  ||
                   isAnchor(workArea, currentPosition, constBACKGROUND)) {
                inProgress = true;
              }
            }
          default  :
        }
        if  (inProgress && (ch != ' ') && (ch != '"') && (ch != '='))
          urlWork.append((char)ch);
        if  (fos != null)  fos.write(ch);
      }
      is.close();
      if  (fos != null)  {
        fos.close();
      }
    } catch(IOException e) {
      try {
        is.close();
        if (fos != null)  fos.close();
      } catch(Exception ee) {}
      try {
        new File(defaultSupplementedFileName).delete();
        success = false;
      } catch(Exception ee) {}
      ControlThread.g.log("# Exception! " + url.toString());
    }
    return success;
  }

  public String defaultSupplement(String file) {
    String temp = ControlThread.g.getLocalPath() + file;
    if (temp.endsWith("/")) {
      return temp + DEFAULTFILENAME;
    }
    return temp;
  }

  public void registerURL(URL url, String str, FileOutputStream fos) throws IOException {
    URL newURL = null;
    if ((str.length() == 0) ||
        (str.indexOf("mailto:") == 0) || 
        (str.indexOf("ftp:")    == 0))   
      return;
    try {
      newURL = new URL(url, str);
    } catch(MalformedURLException e) {
      return;
    }
    String tempString = newURL.toString();
    if  (tempString.indexOf(ControlThread.realm) == 0) {
      ControlThread.pushQueue(tempString);
      if ((tempString.charAt(tempString.length()-1) == '/') && (fos != null)) {
        for (int i = 0; i < DEFAULTFILENAME.length(); i++)
          fos.write(DEFAULTFILENAME.charAt(i));
      }
    }
    return;
  }

  public boolean isAnchor(char[] buffer, int pos, String str) {
    int len = str.length() - 1;
    while(len >= 0)  {
      if (Character.toLowerCase(buffer[pos]) != str.charAt(len))  return false;
      len--;
      if  (--pos < 0)  pos = BUFMAX - 1;
    }
    return true;
  }

  public boolean isHTML(String filename)  {
    return (filename.endsWith(".html") || filename.endsWith(".htm"));
  }

}

